import os
from pyneval.model.swc_node import SwcNode, SwcTree


# if path is a fold
def read_swc_trees(swc_file_paths, tree_name_dict=None):
    """
    Read a swc tree or recursively read all the swc trees in a fold
    Args:
        swc_file_paths(string): path to read swc
        tree_name_dict(dict): a map for swc tree and its file name
            key(SwcTree): SwcTree object
            value(string): name of the swc tree
    Output:
        swc_tree_list(list): a list shaped 1*n, containing all the swc tree in the path
    """
    swc_tree_list = []
    if os.path.isfile(swc_file_paths):
        if not (swc_file_paths[-4:] == ".swc" or swc_file_paths[-4:] == ".SWC"):
            print(swc_file_paths + "is not a swc file")
            return None
        swc_tree = SwcTree()
        swc_tree.load(swc_file_paths)
        swc_tree_list.append(swc_tree)
        if tree_name_dict is not None:
            tree_name_dict[swc_tree] = os.path.basename(swc_file_paths)
    elif os.path.isdir(swc_file_paths):
        for file in os.listdir(swc_file_paths):
            swc_tree = read_swc_trees(swc_file_paths=os.path.join(swc_file_paths, file), tree_name_dict=tree_name_dict)
            if swc_tree is not None:
                swc_tree_list += swc_tree
    return swc_tree_list


def adjust_swcfile(swc_str):
    words = swc_str.split("\n")
    return words


def read_from_str(swc_str):
    swc_tree = SwcTree()
    swc_list = adjust_swcfile(swc_str)
    swc_tree.load_list(swc_list)
    return swc_tree


if __name__ == "__main__":
    tree = SwcTree()
    tree.load()
    pass



demo_func = lambda x: x[0] ** 2 + (x[1] - 0.05) ** 2 + x[2] ** 2

# %% Do SA
from sko.SA import SA

sa = SA(func=demo_func, x0=[1, 1, 0], T_max=1, T_min=1e-9, L=300, max_stay_counter=150)
best_x, best_y = sa.run()
print('best_x:', best_x, 'best_y', best_y)

# %% Plot the result
import matplotlib.pyplot as plt
import pandas as pd

plt.plot(pd.DataFrame(sa.best_y_history).cummin(axis=0))
plt.show()

# %%
from sko.SA import SAFast

sa_fast = SAFast(func=demo_func, x0=[1, 1, 1], T_max=1, T_min=1e-9, q=0.99, L=300, max_stay_counter=150)
sa_fast.run()
print('Fast Simulated Annealing: best_x is ', sa_fast.best_x, 'best_y is ', sa_fast.best_y)

# %%
from sko.SA import SABoltzmann

sa_boltzmann = SABoltzmann(func=demo_func, x0=[1, 1, 1], T_max=1, T_min=1e-9, q=0.99, L=300, max_stay_counter=150)
sa_boltzmann.run()
print('Boltzmann Simulated Annealing: best_x is ', sa_boltzmann.best_x, 'best_y is ', sa_fast.best_y)

# %%
from sko.SA import SACauchy

sa_cauchy = SACauchy(func=demo_func, x0=[1, 1, 1], T_max=1, T_min=1e-9, q=0.99, L=300, max_stay_counter=150)
sa_cauchy.run()
print('Cauchy Simulated Annealing: best_x is ', sa_cauchy.best_x, 'best_y is ', sa_cauchy.best_y)


import os
from anytree import PreOrderIter


def is_path_valid(file_path):
    tmp_path, tmp_file = os.path.split(file_path)

    if not os.path.exists(file_path):
        if not os.path.exists(tmp_path):
            os.makedirs(tmp_path)
        newfile = open(file_path, 'w')
        newfile.close()

    if os.path.isdir(file_path):
        raise Exception("[Error: ] file: \"{}\" has exist in path: \"{}\". and it is a menu"
                        .format(tmp_file, tmp_path))
        return False

    return True


def save_line_tuple_as_swc(match_fail, file_path):
    with open(file_path, 'w') as f:
        f.truncate()
        f.write("# total unmatched edges: {}\n".format(len(match_fail)))
        for line_tuple in match_fail:
            f.write(line_tuple[0].to_swc_str())
        f.write("# --End--\n")


def save_as_swc(object, file_path):
    if not is_path_valid(file_path):
        return False

    if isinstance(object, set):
        save_line_tuple_as_swc(object, file_path)


def print_line_tuple_swc(match_fail_set):
    for line_tuple in match_fail_set:
        print("pos_1: {}, pos_2 {}".format(line_tuple[0]._pos, line_tuple[1]._pos))


def print_swc(object):
    if isinstance(object, set):
        print_line_tuple_swc(object)


def swc_save(swc_tree, out_path, extra=None):
    if not is_path_valid(out_path):
        return False
    swc_node_list = swc_tree.get_node_list()
    swc_tree.sort_node_list(key="id")
    with open(out_path, 'w') as f:
        f.truncate()
        if extra is not None:
            f.write(extra)
        for node in swc_node_list:
            if node.is_virtual():
                continue
            try:
                f.write("{} {} {} {} {} {} {}\n".format(
                    node.get_id(), node._type, node.get_x(), node.get_y(), node.get_z(), node.radius(), node.parent.get_id()
                ))
            except:
                continue



import sys
import jsonschema

from pyneval.model import swc_node
from pyneval.metric.utils import edge_match_utils
from pyneval.io import read_json
from pyneval.io import read_swc
from pyneval.io import swc_writer


def length_metric_run(gold_swc_tree=None, test_swc_tree=None,
                      rad_threshold=-1.0, len_threshold=0.2, debug=False):
    """
    get matched edge set and calculate recall and precision
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        rad_threshold(float): threshold of key point radius
        len_threshold(float): threshold of length of the matching edges
        debug(bool): list debug info ot not
    Returns:
        tuple: contain two values to demonstrate metric result
            precision(float): percentage of total length of edges that are matched compared to test tree
            recall(float): percentage of total length of edges that are matched compared to gold tree
    Raises:
        None
    """
    # get matched edge set
    match_edges, test_match_length = edge_match_utils.get_match_edges(gold_swc_tree=gold_swc_tree,
                                                                      test_swc_tree=test_swc_tree,
                                                                      rad_threshold=rad_threshold,
                                                                      len_threshold=len_threshold,
                                                                      debug=debug)
    # calculate the sum of matched length and total length of gold and test tree
    match_length = 0.0
    for line_tuple in match_edges:
        match_length += line_tuple[0].parent_distance()

    gold_total_length = round(gold_swc_tree.length(), 8)
    test_total_length = round(test_swc_tree.length(), 8)
    match_length = round(match_length, 8)
    test_match_length = round(test_match_length, 8)

    if debug:
        print("match_length = {}, test_match_length = {}, gold_total_length = {}, test_total_length = {}"
              .format(match_length, test_match_length, gold_total_length, test_total_length))
    # calculate recall and precision
    if gold_total_length != 0:
        recall = round(match_length/gold_total_length, 8)
    else:
        recall = 0

    if test_total_length != 0:
        precision = round(test_match_length/test_total_length, 8)
    else:
        precision = 0

    return min(recall, 1.0), min(precision, 1.0)


# @do_cprofile("./mkm_run.prof")
def length_metric(gold_swc_tree, test_swc_tree, config):
    """Main function of length metric.
    unpack config and run the matching function
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            keys: the name of configs
            items: config values
    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\geo_metric_data\\gold_fake_data1.swc")
        test_tree.load("..\\..\\data\\test_data\\geo_metric_data\\test_fake_data1.swc")
        lm_res = length_metric(gold_swc_tree=gold_tree,
                               test_swc_tree=test_tree,
                               config=config)
    Returns:
        tuple: contain two values to demonstrate metric result
            precision(float): percentage of total length of edges that are matched compared to test tree
            recall(float): percentage of total length of edges that are matched compared to gold tree
    Raises:
        None
    """

    # read config
    rad_mode = config["rad_mode"]
    rad_threshold = config["rad_threshold"]
    len_threshold = config["len_threshold"]
    scale = config["scale"]
    debug = config["debug"]

    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)
    gold_swc_tree.set_node_type_by_topo(root_id=1)
    test_swc_tree.set_node_type_by_topo(root_id=5)

    if rad_mode == 1:
        rad_threshold *= -1
    # check every edge in test, if it is overlap with any edge in gold three
    recall, precision = length_metric_run(gold_swc_tree=gold_swc_tree,
                                          test_swc_tree=test_swc_tree,
                                          rad_threshold=rad_threshold,
                                          len_threshold=len_threshold,
                                          debug=debug)

    if "detail_path" in config:
        swc_writer.swc_save(gold_swc_tree, config["detail_path"][:-4]+"_gold.swc")
        swc_writer.swc_save(test_swc_tree, config["detail_path"][:-4]+"_test.swc")
    if debug:
        print("Recall = {}, Precision = {}".format(recall, precision))

    res = {
        "recall": recall,
        "precision": precision
    }
    return res, gold_swc_tree, test_swc_tree


def web_length_metric(gold_swc, test_swc, mode, rad_threshold, len_threshold):
    """
        length metric interface connect to webmets
        (https://github.com/bennieHan/webmets)
        which is not used nowadays
    Args:
        gold_swc(SwcTree):
        test_swc(SwcTree):
        mode(int) 1 or 2, same as "mode" in config
        rad_threshold(float): threshold of key point radius
        len_threshold(float): threshold of length of the matching edges
    Example:
        # in Flask:
        def geometry():
            gold_swc = request.form.get('gold_txt')
            test_swc = request.form.get('test_txt')
            method = request.form.get('method')
            rad_threshold = request.form.get('rad_threshold')
            len_threshold = request.form.get('len_threshold')
            result = web_length_metric(gold_swc=gold_swc,
                                       test_swc=test_swc,
                                       method=int(method),
                                       rad_threshold=rad_threshold,
                                       len_threshold=len_threshold)
            # result['gold_swc'] = adjust_radius_work(result['gold_swc'], 0.1)
            # result['test_swc'] = adjust_radius_work(result['test_swc'], 0.1)
            return jsonify(result=result)
    Returns:
        tuple: contain five values to demonstrate metric result
            precision(float): percentage of total length of edges that are matched compared to test tree
            recall(float): percentage of total length of edges that are matched compared to gold tree
            gold_swc(Swc Tree):
            test_swc(Swc Tree):
            vertical_swc(Swc Tree): vertical edge between gold and test swc tree.
    Raises:
        None
    """
    gold_tree = swc_node.SwcTree()
    test_tree = swc_node.SwcTree()

    gold_tree.load_list(read_swc.adjust_swcfile(gold_swc))
    test_tree.load_list(read_swc.adjust_swcfile(test_swc))

    config = {
        'method': mode,
        'len_threshold': len_threshold,
        'rad_threshold': rad_threshold
    }

    lm_res = length_metric(gold_swc_tree=gold_tree,
                           test_swc_tree=test_tree,
                           config=config)

    result = {
        'recall': lm_res[0],
        'precision': lm_res[1],
        'gold_swc': gold_tree.to_str_list(),
        'test_swc': test_tree.to_str_list(),
        'vertical_swc': lm_res[2]
    }
    return result


if __name__ == "__main__":
    goldTree = swc_node.SwcTree()
    testTree = swc_node.SwcTree()
    sys.setrecursionlimit(10000000)
    testTree.load("E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\data\example_selected\\a.swc")
    goldTree.load("E:\\00_project\\00_neural_reconstruction\\01_project\PyNeval\output\\random_data\move\\a\\020\move_00.swc")

    config = read_json.read_json("..\\..\\config\\length_metric.json")
    config_schema = read_json.read_json("..\\..\\config\\schemas\\length_metric_schema.json")
    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")
    config["detail_path"] = "..\\..\\output\\length_output\\length_metric_detail.swc"
    
    lm_res = length_metric(gold_swc_tree=goldTree,
                           test_swc_tree=testTree,
                           config=config)

    print("recall    = {}\n"
          "precision = {}\n".format(lm_res["recall"], lm_res["precision"]))



from pyneval.metric.utils.klib.TiffFile import imsave, imread, TiffFile
import os


def read_tiffs(tiff_paths):
    tiff_list = []
    if os.path.isfile(tiff_paths):
        if not (tiff_paths[-4:] == ".tif" or tiff_paths[-4:] == ".TIF"):
            print(tiff_paths + "is not a tif file")
            return None
        tiff_file = imread(tiff_paths)
        tiff_list.append(tiff_file)
    elif os.path.isdir(tiff_paths):
        for file in os.listdir(tiff_paths):
            tiff_list += read_tiffs(tiff_paths=os.path.join(tiff_paths, file))
    return tiff_list


# -*- coding: utf-8 -*-
"""a geometry metric method, SSD metric

This module implement a geometry metric method called Substantial Spatial Distance.
relative paper: https://doi.org/10.1038/nbt.1612
title, auther

Example:
    In command line:
    $ pyneval --gold ./data\\test_data\\ssd_data\\gold\\c.swc
              --test .\\data\\test_data\\ssd_data\\test\\c.swc
              --metric ssd_metric
Attributes:
    None

Todos:
    None
"""
import sys

import jsonschema

from pyneval.io import swc_writer
from pyneval.model import swc_node
from pyneval.io import read_json
from pyneval.tools import re_sample
from pyneval.metric.utils import point_match_utils
from pyneval.io import read_config


def get_mse(src_tree, tar_tree, ssd_threshold=2.0, mode=1):
    """ calculate the minimum square error of two trees
    find the closest node on the tar_tree for each node on the src tree, calculate the average
    distance of these node pairs.

    Args:
        src_tree(SwcTree):
        tar_tree(SwcTree):
        ssd_threshold(float): distance will be count into the res
            only if two nodes' distance is larger than this threshold.
        mode(1 or 2):
            1 means static threshold, equal to ssd_threshold.
            2 means dynamic threshold, equal to ssd_threshold * src_node.threshold

    Returns:
        dis(float): average minimum distance of node, distance must be larger than ssd_threshold
        num(float): The number of pairs of nodes that are counted when calculating distance

    Raise:
        None
    """
    dis, num = 0, 0
    kdtree, pos_node_dict = point_match_utils.create_kdtree(tar_tree.get_node_list())
    for node in src_tree.get_node_list():
        if node.is_virtual():
            continue
        target_pos = kdtree.search_knn(list(node.get_center_as_tuple()), k=1)[0]
        target_node = pos_node_dict[tuple(target_pos[0].data)]

        cur_dis = target_node.distance(node)

        if mode == 1:
            threshold = ssd_threshold
        else:
            threshold = ssd_threshold * node.radius()

        if cur_dis >= threshold:
            node._type = 9
            dis += cur_dis
            num += 1
    try:
        dis /= num
    except ZeroDivisionError:
        dis = num = 0
    return dis, num


def ssd_metric(gold_swc_tree: swc_node.SwcTree, test_swc_tree: swc_node.SwcTree, config: dict):
    """Main function of SSD metric.
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            The keys of 'config' is the name of configs, and the items are config values
    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\ssd_data\\gold\\c.swc")
        test_tree.load("..\\..\\data\\test_data\\ssd_data\\test\\c.swc")
        score, recall, precision = ssd_metric(gold_swc_tree=gold_tree,
                                              test_swc_tree=test_tree,
                                              config=config)
    Returns:
        tuple: contain three values to demonstrate metric result
            avg_score(float): average distance of nodes between gold and test swc trees.
            precision(float): percentage of nodes that are matched compared to test tree
            recall(float): percentage of nodes that are matched compared to gold tree

    Raises:
        None
    """
    debug = config["debug"]
    threshold_mode = config["threshold_mode"]
    ssd_threshold = config["ssd_threshold"]
    up_sample_threshold = config["up_sample_threshold"]
    scale = config["scale"]

    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)
    u_gold_swc_tree = re_sample.up_sample_swc_tree(swc_tree=gold_swc_tree,
                                                   length_threshold=up_sample_threshold)
    u_test_swc_tree = re_sample.up_sample_swc_tree(swc_tree=test_swc_tree,
                                                   length_threshold=up_sample_threshold)
    u_gold_swc_tree.set_node_type_by_topo(root_id=1)
    u_test_swc_tree.set_node_type_by_topo(root_id=5)

    g2t_score, g2t_num = get_mse(src_tree=u_gold_swc_tree, tar_tree=u_test_swc_tree,
                                 ssd_threshold=ssd_threshold, mode=threshold_mode)
    t2g_score, t2g_num = get_mse(src_tree=u_test_swc_tree, tar_tree=u_gold_swc_tree,
                                 ssd_threshold=ssd_threshold, mode=threshold_mode)

    if debug:
        print("recall_num = {}, pre_num = {}, gold_tot_num = {}, test_tot_num = {} {} {}".format(
            g2t_num, t2g_num, u_gold_swc_tree.size(), u_test_swc_tree.size(), gold_swc_tree.length(), test_swc_tree.length()
        ))

    res = {
        "avg_score": (g2t_score + t2g_score) / 2,
        "recall": 1 - g2t_num/u_gold_swc_tree.size(),
        "precision": 1 - t2g_num/u_test_swc_tree.size()
    }

    if "detail" in config:
        return res, u_gold_swc_tree, u_test_swc_tree
    return res


if __name__ == "__main__":
    test_tree = swc_node.SwcTree()
    gold_tree = swc_node.SwcTree()

    sys.setrecursionlimit(10000000)
    gold_tree.load("../../data/test_data/geo_metric_data/gold_fake_data1.swc")
    test_tree.load("../../data/test_data/geo_metric_data/test_fake_data1.swc")

    config = read_json.read_json("../../config/ssd_metric.json")
    config_schema = read_json.read_json("../../config/schemas/ssd_metric_schema.json")

    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")
    config["detail_path"] = "..//..//output//ssd_output//ssd_detail.swc"

    ssd_res = ssd_metric(gold_swc_tree=gold_tree,
                         test_swc_tree=test_tree,
                         config=config)

    print("ssd score = {}\n"
          "recall    = {}%\n"
          "precision = {}%".
          format(round(ssd_res["avg_score"], 2),
                 round(ssd_res["recall"]*100, 2),
                 round(ssd_res["precision"]*100, 2)))




import sys
import jsonschema

from pyneval.io import read_json
from pyneval.model import swc_node
from pyneval.metric.utils import km_utils
from pyneval.metric.utils import point_match_utils


def get_result(test_len, gold_len, switch, km, threshold_dis):
    false_pos_num, false_neg_num, true_pos_num = 0, 0, 0
    # count numer of nodes which are matched, calculate FP, TN, TP
    for i in range(0, gold_len):
        if km.match[i] != -1 and km.G[km.match[i]][i] != -0x3f3f3f3f / 2:
            true_pos_num += 1
    false_neg_num = gold_len - true_pos_num
    false_pos_num = test_len - true_pos_num

    # definition of swich is in function "get_dis_graph"
    if switch:
        false_neg_num, false_pos_num = false_pos_num, false_neg_num

    if true_pos_num != 0:
        mean_dis = -km.get_max_dis() / true_pos_num
    else:
        mean_dis = 0.0
    if mean_dis == -0.0:
        mean_dis = 0.0

    pt_cost = -km.get_max_dis() + threshold_dis * (false_neg_num + false_pos_num) / (
                false_neg_num + false_pos_num + true_pos_num)

    # debug:
    # print("output")
    # print(false_pos_num)
    # print(true_neg_num)
    # print(mean_dis)
    # print(pt_cost)
    return true_pos_num, false_neg_num, false_pos_num, mean_dis, mean_dis * true_pos_num, pt_cost


def get_colored_tree(test_node_list, gold_node_list, switch, km, color):
    '''
    color[0] = tp's color
    color[1] = fp's color
    color[2] = fn's color
    '''
    tp_set = set()
    for i in range(0, len(gold_node_list)):
        if km.match[i] != -1 and km.G[km.match[i]][i] != -0x3f3f3f3f / 2:
            gold_node_list[i]._type = color[0]
            test_node_list[km.match[i]]._type = color[0]
            tp_set.add(test_node_list[km.match[i]])
        else:
            if switch:
                gold_node_list[i]._type = color[1]
            else:
                gold_node_list[i]._type = color[2]
    for node in test_node_list:
        if node not in tp_set:
            if switch:
                node._type = color[2]
            else:
                node._type = color[1]


def score_point_distance(gold_tree: swc_node.SwcTree, test_tree: swc_node.SwcTree,
                         test_node_list: list, gold_node_list: list,
                         threshold_dis: float, color: list):
    """
    get minimum matching distance by running KM algorithm
    than calculte the return value according to matching result
    Args:
        gold_tree(Swc Tree)
        test_tree(Swc Tree)
        gold_node_list(List): contains only branch nodes
        test_node_list(List): contains only branch nodes
        threshold_dis: if the distance of two node are larger than this threshold,
                       they are considered unlimited far
        color(List): color id of tp, fn, fp nodes
    Returns:
        gold_len(int): length of gold_node_list
        test_len(int): length of test_node_list
        true_pos_num(int): number of nodes in both gold and test tree
        false_neg_num(int): number of nodes in gold but not test tree
        false_pos_num(int): number of nodes in test but not gold tree
        mean_dis: mean distance of nodes that are successfully matched(true positive)
        tot_dis: total distance of nodes that are successfully matched(true positive)
        pt_cost: a composite value calculated by tp, fn, fp and threshold
        iso_node_num: number of nodes in test tree without parents or children
    """
    test_gold_dict = point_match_utils.get_swc2swc_dicts(src_node_list=test_tree.get_node_list(),
                                                         tar_node_list=gold_tree.get_node_list())
    # disgraph is a 2D ndarray store the distance between nodes in gold and test
    # test_node_list contains only branch or leaf nodes
    dis_graph, switch, test_len, gold_len = km_utils.get_dis_graph(gold_tree=gold_tree,
                                                                   test_tree=test_tree,
                                                                   test_node_list=test_node_list,
                                                                   gold_node_list=gold_node_list,
                                                                   test_gold_dict=test_gold_dict,
                                                                   threshold_dis=threshold_dis,
                                                                   metric_mode=1)
    # create a KM object and calculate the minimum match
    km = km_utils.KM(maxn=max(test_len, gold_len)+10, nx=test_len, ny=gold_len, G=dis_graph)
    km.solve()
    # calculate the result
    true_pos_num, false_neg_num, false_pos_num, \
    mean_dis, tot_dis, pt_cost = get_result(test_len=test_len,
                                            gold_len=gold_len,
                                            switch=switch,
                                            km=km,
                                            threshold_dis=threshold_dis)
    # calculate the number of isolated nodes
    iso_node_num = 0
    for node in test_tree.get_node_list():
        if node.is_isolated():
            iso_node_num += 1
    # get a colored tree with
    get_colored_tree(test_node_list=test_node_list, gold_node_list=gold_node_list,
                     switch=switch, km=km, color=color)
    return gold_len, test_len, true_pos_num, false_neg_num, false_pos_num, \
           mean_dis, tot_dis, pt_cost, iso_node_num


def branch_leaf_metric(gold_swc_tree, test_swc_tree, config):
    """
    branch metric calculates the minimum distance match between branches of two swc trees
    This function is used for unpacking configs and packaging return values
    Args:
        gold_swc_tree(Swc Tree) gold standard tree
        test_swc_tree(Swc Tree) reconstructed tree
        config(dict):
            keys: the name of configs
            items: config values
    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\topo_metric_data\\gold_fake_data1.swc")
        test_tree.load("..\\..\\data\\test_data\\topo_metric_data\\test_fake_data1.swc")
        branch_result = length_metric(gold_swc_tree=gold_tree,
                                      test_swc_tree=test_tree,
                                      config=config)
        print(branch_result["mean_dis"])
        ...
    Return:
         branch_result(tuple) a tuple of 9 metric results
    """
    # read configs
    threshold_dis = config["threshold_dis"]
    threshold_mode = config["threshold_mode"]
    scale = config["scale"]

    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)
    # in threshold mode 2, threshold is a multiple of the average length of edges
    if threshold_mode == 2:
        # length of the entire gold swc forest
        tot_dis = gold_swc_tree.length()
        # number of edges in the forest
        edge_num = len(gold_swc_tree.get_node_list())-1-len(gold_swc_tree.root().children)
        threshold_dis = threshold_dis * tot_dis / edge_num
    # denote the color id of different type of nodes.
    color = [
        config["true_positive"],
        config["missed"],
        config["excess"]
    ]
    gold_swc_tree.type_clear(0, 0)
    test_swc_tree.type_clear(0, 0)
    test_branch_swc_list = test_swc_tree.get_branch_swc_list()
    gold_branch_swc_list = gold_swc_tree.get_branch_swc_list()

    branch_result_tuple = score_point_distance(gold_tree=gold_swc_tree,
                                               test_tree=test_swc_tree,
                                               test_node_list=test_branch_swc_list,
                                               gold_node_list=gold_branch_swc_list,
                                               threshold_dis=threshold_dis,
                                               color=color)

    branch_result = {
        "gold_len": branch_result_tuple[0],
        "test_len": branch_result_tuple[1],
        "true_pos_num": branch_result_tuple[2],
        "false_neg_num": branch_result_tuple[3],
        "false_pos_num": branch_result_tuple[4],
        "mean_dis": branch_result_tuple[5],
        "tot_dis": branch_result_tuple[6],
        "pt_cost": branch_result_tuple[7],
        "iso_node_num": branch_result_tuple[8]
    }
    return branch_result, gold_swc_tree, test_swc_tree


if __name__ == "__main__":
    sys.setrecursionlimit(1000000)
    file_name = "fake_data11"
    gold_swc_tree = swc_node.SwcTree()
    test_swc_tree = swc_node.SwcTree()

    gold_swc_tree.load("../../data/example_selected/a.swc")
    test_swc_tree.load("../../output/random_data/move/a/010/move_03.swc")

    config = read_json.read_json("..\\..\\config\\branch_metric.json")
    config_schema = read_json.read_json("..\\..\\config\\schemas\\branch_metric_schema.json")
    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")

    branch_result = \
        branch_leaf_metric(test_swc_tree=test_swc_tree, gold_swc_tree=gold_swc_tree, config=config)
    print("---------------Result---------------")
    print("gole_branch_num = {}, test_branch_num = {}\n"
          "true_positive_number  = {}\n"
          "false_negative_num    = {}\n"
          "false_positive_num    = {}\n"
          "matched_mean_distance = {}\n"
          "matched_sum_distance  = {}\n"
          "pt_score              = {}\n"
          "isolated node number  = {}".
          format(branch_result["gold_len"], branch_result["test_len"], branch_result["true_pos_num"],
                 branch_result["false_neg_num"], branch_result["false_pos_num"], branch_result["mean_dis"],
                 branch_result["tot_dis"], branch_result["pt_cost"], branch_result["iso_node_num"]))
    print("----------------End-----------------")
    # with open("../../output/branch_metric/{}_gold.swc".format(file_name), 'w') as f:
    #     f.write(gold_swc_tree.to_str_list())
    # with open("../../output/branch_metric/{}_test.swc".format(file_name), 'w') as f:
    #     f.write(test_swc_tree.to_str_list())




DINF = 999999999999.0
EPS = 0.0000001
SAME_POS_TH = 0.03
MIN_SIZE = 0.8
FLOAT_ERROR = 0.001

def get_default_threshold(gold_swc_tree):
    total_length = gold_swc_tree.length()
    total_node = gold_swc_tree.size()
    if total_node <= 1:
        dis_threshold = 0.1
    else:
        dis_threshold = (total_length/total_node)/10
    return dis_threshold



from setuptools import setup
import glob

setup(name='pyneval',
      version='0.0.1',
      description='pyneval, a toolbox for quantitative metric on reconstructed neuron',
      author='HanZhang, TingZhao',
      author_email='benniehan98@gmail.com',
      url='https://github.com/bennieHan/PyNeval.git',
      packages=['pyneval', 'pyneval.cli', 'test', 'test.test_model',
                'test.test_model.diadem_metric', 'test.test_model.length_metric',
                'pyneval.io', 'pyneval.metric', 'pyneval.metric.utils', 'pyneval.metric.utils.klib', 'pyneval.model',
                'pyneval.tools', 'pyneval.tools.optimize'],
      py_modules=['pyneval', 'pyneval.cli', 'test', 'test.test_model',
                'test.test_model.diadem_metric', 'test.test_model.length_metric',
                'pyneval.io', 'pyneval.metric', 'pyneval.metric.utils', 'pyneval.metric.utils.klib', 'pyneval.model',
                'pyneval.tools', 'pyneval.tools.optimize'],
      data_files=[('./lib/site-packages/config', glob.glob('config/*.json')),
                  ("./lib/site-packages/config/schemas", glob.glob('config/schemas/*.json'))],
      install_requires=[
            'anytree>=2.7.2',
            'kdtree>=0.16',
            'numpy>=1.0',
            'rtree>=0.8.0',
            'jsonschema>=3.2.0'
      ],
      entry_points={
          'console_scripts': [
              'pyneval=pyneval.cli.pyneval:run',
              'pyneval_tools=pyneval.cli.pyneval_tools:run',
          ]
      }
)




import jsonschema
import queue
import math

from pyneval.io import read_json
from pyneval.io.swc_writer import swc_save
from pyneval.metric.utils.klib.TiffFile import imread
from pyneval.metric.utils.tiff_utils import front_expend_step
from pyneval.model.swc_node import SwcTree
from pyneval.tools.re_sample import up_sample_swc_tree


def adjust_swc_tree(swc_tree, tiff_file, thres_intensity, max_step=4):
    swc_list = swc_tree.get_node_list()
    for node in swc_list:
        if node.is_virtual():
            continue
        new_center = cal_label(node=node, test_tiff=tiff_file,
                               thres_intensity=thres_intensity, max_step=max_step)
        if not new_center:
            print("{} {} {} ".format(
                node.get_x(), node.get_y(), node.get_z()
            ))
            print("[Warning: ] step too little")
        else:
            node.set_x(new_center[0])
            node.set_y(new_center[1])
            node.set_z(new_center[2])


def get_dis(tuple_1, tuple_2):
    dx = tuple_1[0] - tuple_2[0]
    dy = tuple_1[1] - tuple_2[1]
    dz = tuple_1[2] - tuple_2[2]
    return math.sqrt(dx**2 + dy**2 + dz**2)


def cal_label(node, test_tiff, thres_intensity, max_step=0):
    que = queue.Queue()
    center = [round(node.get_x()), round(node.get_y()), round(node.get_z())]
    que.put(tuple([center, 0]))
    vis = set()
    vis.add(tuple(center))

    stp = [
        [1, 0, 0], [0, 1, 0], [0, 0, 1],
        [-1, 0, 0], [0, -1, 0], [0, 0, -1]
    ]

    while not que.empty():
        cur = que.get()
        cur_loc = cur[0]
        cur_step = cur[1]
        if cur_step > max_step:
            continue

        if 0 <= round(cur_loc[0]) < test_tiff.shape[2] and\
           0 <= round(cur_loc[1]) < test_tiff.shape[1] and\
           0 <= round(cur_loc[2]) < test_tiff.shape[0] and\
           test_tiff[round(cur_loc[2])][round(cur_loc[1])][round(cur_loc[0])] >= thres_intensity:
            return tuple([round(cur_loc[0]), round(cur_loc[1]), round(cur_loc[2])])

        for i in range(6):
            dx = cur_loc[0] + stp[i][0]
            dy = cur_loc[1] + stp[i][1]
            dz = cur_loc[2] + stp[i][2]
            new_pos = tuple([dx, dy, dz])
            if new_pos not in vis:
                vis.add(tuple([new_pos, cur_step + 1]))
                que.put(tuple([new_pos, cur_step + 1]))
    return None


def cal_volume_recall(test_tiff, gold_swc, intensity_threshold, debug):
    swc_node_list = gold_swc.get_node_list()
    tot_front, tot_back = 0, 0

    for node in swc_node_list:
        if node.is_virtual():
            continue
        if (len(node.children) == 0 or node.parent.is_virtual()) and\
                cal_label(node, test_tiff, intensity_threshold, 0) is not None:
            tot_front += 1
        elif cal_label(node, test_tiff, intensity_threshold, 1) is not None:
            tot_front += 1
        else:
            if debug:
                print("[Info: ] fail: {} {} {} {} {}".format(
                    node.get_id(), round(node.get_x()), round(node.get_y()), round(node.get_z()),
                    test_tiff[round(node.get_z())][round(node.get_y())][round(node.get_x())]
                ))

        tot_back += 1
    return tot_front/tot_back


def volume_metric(gold_swc_tree, test_swc_tree, config=None):
    length_threshold = config['length_threshold']
    intensity_threshold = config['intensity_threshold']
    debug = config['debug']
    densed_swc_tree = up_sample_swc_tree(swc_tree=gold_swc_tree, length_threshold=length_threshold)
    recall = cal_volume_recall(test_swc_tree, densed_swc_tree, intensity_threshold, debug)

    res = {
        "recall": recall
    }
    return res, None, None


if __name__ == "__main__":
    swc_path = "..\\..\\data\example_selected\g.swc"
    tiff_path = "..\\..\\data\example_selected\g.tif"
    test_tiff = imread(tiff_path)
    swc_tree = SwcTree()
    swc_tree.load(swc_path)
    config = read_json.read_json("..\\..\\config\\volume_metric.json")
    config_schema = read_json.read_json("..\\..\\config\\schemas\\volume_metric_schema.json")

    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")

    res = volume_metric(tiff_test=test_tiff, swc_gold=swc_tree, config=config)
    print("recall = {}".format(res["recall"]))
